/*
 * csmngr.h - header file for csmngr.
 *
 * Copyright (C) 2024 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See LICENSE file for license related information.
 *
 */

#ifndef CSMNGR_H
#define CSMNGR_H

#include <wifi.h>

#ifdef HAS_UBUS
#include <libubus.h>
#endif

#include "useropts.h"

#define DEFAULT_HISTORY_FILE            "/etc/wifi_acs_history.json"
#define DEFAULT_HISTORY_TMPFS_FILE      "/tmp/wifi_acs_history.json"
#define DEFAULT_HISTORY_AGEOUT          604800		/* 7 days in seconds */
#define DEFAULT_HISTORY_MAX_NUM         14		/* max number of history records */
#define DEFAULT_HISTORY_MAX_SIZE        0x100000	/* 1Mb */

#define OBJECT_INVALID	((uint32_t)-1)

#define CSMNGR_EVENT			"acs"
#define CSMNGR_EVENT_SURVEY_START	"acs-start"
#define CSMNGR_EVENT_SURVEY_STOP	"acs-stop"


#define BAND2_NOISE_FLOOR		-101	/* in dBm */
#define BAND5_NOISE_FLOOR		-98
#define BAND6_NOISE_FLOOR		-99

#define RADIO_IDLE_THRESHOLD		15.0	/* in % */

/* all the intervals and timeouts below are in seconds, unless mentioned */
#define CSMNGR_DEFAULT_REFRESH_INT	43200	/* 12 hrs auto-refresh cycle */
#define CSMNGR_DEFER_REFRESH_INT	7200	/* 2 hours */

#define CAC_TIMEOUT			60
#define CAC_TIMEOUT_WRDR		600

#define SCAN_FINISHED_TIMEOUT		10	/* per-channel */
#define SCAN_FAIL_RETRY_INT		20
#define CAC_FAIL_RETRY_INT		65

#define CSMNGR_DEFAULT_COLLECT_INT	5

#define MAX_NUM_EXCLUDE_OPCLASS		32
#define MAX_NUM_EXCLUDE_CHANNELS	32
#define MAX_NUM_EXCLUDE_BANDWIDTHS	8

#define MAX_NUM_PREF_CHANNELS		32

enum csmngr_mode {
	CS_MODE_DISABLED,			/* disabled */
	CS_MODE_ENABLED,
	CS_MODE_ACS,				/* enabled + select channel */
	CS_MODE_MONITOR,			/* enabled + decide channel */
	CS_MODE_DRY_RUN,			/* debug mode */
};

enum csmngr_state {
	CS_STATE_IDLE,			/* doing nothing */
	CS_STATE_SCANNING,		/* radio scanning any channel */
	CS_STATE_PROCESS,		/* processing survey data */
	CS_STATE_CAC,			/* CAC ongoing */
	CS_STATE_END,
};

struct csmngr_config {
	const char *objname;
	bool enabled;
	bool history;			/* keep history */
	bool history_persist;		/* keep history across reboot */
#define FILE_PATH_MAX	512
	char history_file[FILE_PATH_MAX];
	char hosts_file[FILE_PATH_MAX];
	uint32_t history_ageout;	/* ageout history entry in seconds */
	uint32_t history_maxnum;	/* max number of history entries */
	uint32_t history_maxsize;	/* max size of history entries (k/m) */
};

#if 0
struct cac_preclear {
	uint32_t start_freq;
	uint32_t bandwidth;
	struct list_head list;
};
#endif

struct cac_preclear {
	uint32_t channel;
	enum wifi_bw bw;
	uint8_t mode;
};

struct radio_2ghz {
	bool pick_1_6_11;
};

struct radio_5ghz {
	//struct list_head caclist;
	struct cac_preclear caclist[16];
	int next_cac;
	size_t num_bgcac;
	bool bgdfs;
	bool tpc;
};

struct radio_6ghz {
	char puncture_bitmap[128];
	uint8_t puncture_threshold;	/* in %ge above which punctured */
};

struct channel_info_entry {
	time_t tsp;
	struct chan_entry info;
	struct list_head list;
};

struct channel_info {
	uint32_t freq;
	uint32_t channel;
	int num;		/* num struct channel_info_entry */
	struct list_head chinfolist;
};

struct channel_survey_data {
	void *data;		/* points to array of struct chan_entry */
	int num;		/* num chan_entry */
	time_t tsp;		/* latest time of a survey entry in this group */
	struct list_head list;
};

#define NUM_BANDWIDTHS		5	/* 20, 40, 80, 160 and 320 */

enum cs_criteria_type {
	INTFX_FACTOR,		/* interference factor */
	NUM_OBSS,		/* number of overlapping BSS */
	REG_TXPOWER,		/* regulatory txpower */
	NUM_CS_CRITERIA,
};

#define CS_INTFX_FACTOR_WEIGHT	60
#define CS_NUM_OBSS_WEIGHT	30
#define CS_REG_TXPOWER		10

struct csmngr_verdict_entry {
	uint32_t opclass;
	uint32_t channel;	/* ctrl-channel within opclass */
	enum wifi_bw bw;	//TODO: remove redundant
	double score;		/* total weighted-score from following criteria */
	struct cs_criteria {
		uint8_t data[64];
		uint8_t wtdata[64];	/* weighted data */
	} opt[NUM_CS_CRITERIA];
};

struct csmngr_verdict {
	uint8_t opclass;	/* best opclass in a bandwidth category */
	enum wifi_bw bw;	/* redundant info (for debug only) */
	uint32_t channel;	/* control-channel in opclass */
	double score;		/* total weighted-score from following criteria */
	int nent;		/* num ent */
	struct csmngr_verdict_entry ent[64];	//TODO: optimize size
};

#define MAX_NUM_SUPP_CHANNELS		64
#define MAX_NUM_PER_CHANNEL_SCANRES	64

/* represent a 20MHz channel information */
struct channel_data {
	uint32_t channel;		/* 20MHz ctrl-channel */
	double intfx;			/* interference factor */
	int num_scanresults;
	struct wifi_bss scanresults[MAX_NUM_PER_CHANNEL_SCANRES];
};

struct radio_interface {
	bool enabled;
	char ifname[16];
	int ifindex;
	uint8_t macaddr[6];
	uint32_t ifflags;
	uint8_t operstate;
	bool is_ap;
	bool is_brif;
	bool has_bsta;			/* whether hosts a bsta interface */
	uint32_t brport;
	int br_ifindex;			/* master ifindex if enslaved to bridge */
	int nomaster;			/* is set to 1 when not enslaved */
	int carrier;			/* 1 = on, 0 = off */

	char country[4];		/* country code */
	char phyname[256];
	bool multiband;
	enum wifi_band band;
	struct wifi_radio info;
	uint8_t oper_std;			/* current operating std */
	struct wifi_opclass curr_opclass;	/* current opclass */
	int num_supp_opclass;
	struct wifi_opclass supp_opclass[64];	/* supported opclasses */
	uint32_t supp_bandwidths;		/* bitmap of supported BW* */

	/* supp-opclass - pref-opclass = usable-opclass */
	int num_usable_opclass;
	struct wifi_opclass usable_opclass[64];

	union {
		struct radio_6ghz radio6;
		struct radio_5ghz radio5;
		struct radio_2ghz radio2;
	};

	uint32_t idle_threshold;
	uint32_t refresh_int;		/* refresh interval in seconds */
	uint32_t collect_int;		/* collection interval in seconds */
	uint32_t collect_dur;		/* duration in msecs */

	atimer_t refreshtimer;
	atimer_t collecttimer;
	atimer_t runtimer;

	bool include;

	enum csmngr_state state;
	int refresh;

	/* survey data */
	int num_chinfo;
	struct chan_entry chinfo[64];
	time_t tsp[64];
	double intfx[64];		/* interference factor */

	struct channel_data chdata[MAX_NUM_SUPP_CHANNELS];

	uint32_t max_bandwidth;		//TODO: fill

	int num_survey;			/* num channel_survey_data entries */
	struct list_head surveylist;	/* list of channel_survey_data */

	bool cac_ongoing;

	int ncac;
	uint32_t scan_channel;
	uint32_t scantimeout;		/* per-channel scan timeout */
	int nscan;
	int num_supp_channels;
	uint32_t supp_channels[64];	/* 20MHz channels supported */

	enum csmngr_mode mode;
	uint8_t exclude_opclass[MAX_NUM_EXCLUDE_OPCLASS];
	uint8_t exclude_channels[MAX_NUM_EXCLUDE_CHANNELS];
	uint8_t exclude_bandwidths[MAX_NUM_EXCLUDE_BANDWIDTHS];
	int num_exclude_bandwidth;

	bool exclude_dfs;
	bool exclude_non_psc;

	bool autochannel;	/* true when channel=auto in wireless config */

	int num_pref_channels;
	uint8_t pref_channels[MAX_NUM_PREF_CHANNELS];
	uint8_t pref_channel_bias[MAX_NUM_PREF_CHANNELS];

	int8_t nf;		/* noise floor in dBm */

	/* override channel and bandwidth */
	uint8_t channel;
	uint32_t bandwidth;

	/* weights of cs criteria: -1 to ignore, else 0 - 100 in % */
	int8_t cs_attr_weight[NUM_CS_CRITERIA];

	struct csmngr_verdict verdict[NUM_BANDWIDTHS];

	void *priv;
	struct list_head list;
};

struct csmngr_private {
	struct csmngr_useropts uopts;
	struct csmngr_config cfg;
	atimer_t hbtimer;

	int num_radio;
	struct list_head radiolist;	/* list of struct radio_interface */

	int cmdsock;
	char *cmdpath;
	struct uloop_fd cmdloop;

#ifdef HAS_UBUS
	struct ubus_context *bus;
	struct ubus_object obj;
	struct ubus_event_handler evh;
#endif
};


uint32_t wifi_bw2mhz(enum wifi_bw bw);
enum wifi_bw wifi_mhz2bw(int b);

int csmngr_init(void **priv, void *user_opts);
int csmngr_exit(void *priv);
void csmngr_run(void *priv);


int csmngr_start_engine(struct csmngr_private *priv);


int csmngr_radio_scan(struct radio_interface *r, uint32_t channel);
int csmngr_radio_collect_channels_info(struct radio_interface *r);
int csmngr_radio_collect_scanresults(struct radio_interface *r);

int csmngr_radio_run(struct radio_interface *r);
void csmngr_radio_runtimer_cb(atimer_t *t);

int csmngr_process_event_radio_scan_started(struct csmngr_private *priv,
					    struct radio_interface *r);

int csmngr_process_event_radio_scan_finished(struct csmngr_private *priv,
					     struct radio_interface *r);

int csmngr_process_event_radio_cac_start(struct csmngr_private *priv,
					 struct radio_interface *r,
					 uint32_t channel, uint32_t bandwidth,
					 uint32_t cac_time);

int csmngr_process_event_radio_cac_end(struct csmngr_private *priv,
				       struct radio_interface *r,
				       uint32_t channel, uint32_t bandwidth,
				       uint32_t status);

int csmngr_process_event_radio_radar(struct csmngr_private *priv,
				     struct radio_interface *r,
				     uint32_t channel, uint32_t bandwidth);

int csmngr_start_cmdserver(struct csmngr_private *priv, struct csmngr_useropts *u);

int csmngr_status(struct csmngr_private *priv, void *out);
int csmngr_results(struct csmngr_private *priv, void *out);
int csmngr_refresh(struct csmngr_private *priv, void *args, void *out);
int csmngr_dump_history(struct csmngr_private *priv, void *out);
int csmngr_include_radio(struct csmngr_private *priv, const char *ifname, bool include);
int csmngr_add_radio(struct csmngr_private *priv, void *args, void *out);
int csmngr_del_radio(struct csmngr_private *priv, void *args, void *out);
int csmngr_set(struct csmngr_private *priv, void *args, void *out);
int csmngr_scan(struct csmngr_private *priv, void *args, void *out);
int csmngr_dump_survey(struct csmngr_private *priv, void *args, void *out);

int csmngr_do_channel_switch(struct csmngr_private *priv, void *args, void *out);
int csmngr_channel_switch(struct radio_interface *r, uint32_t opclass,
			  uint32_t channel, uint32_t bandwidth);

int csmngr_config_defaults(struct csmngr_private *priv);
int csmngr_dump_config(struct csmngr_private *priv);
int csmngr_reconfig(struct csmngr_private *priv, const char *path, const char *file);
void csmngr_config_free(struct csmngr_private *priv);

struct radio_interface *csmngr_alloc_radio(struct csmngr_private *priv, const char *ifname);
struct radio_interface *csmngr_lookup_radio(struct csmngr_private *priv, const char *ifname);
struct radio_interface *csmngr_lookup_radio_by_band(struct csmngr_private *priv, uint8_t band);
struct radio_interface *csmngr_lookup_radio_by_interface(struct csmngr_private *priv, const char *ifname);
char *csmngr_radio_get_first_ap_ifname(struct radio_interface *r);
char *csmngr_radio_get_first_sta_ifname(struct radio_interface *r);

struct channel_data *radio_lookup_channel_data(struct radio_interface *r, uint32_t channel);
int radio_get_chinfo_index_by_channel(struct radio_interface *r, uint32_t channel);	//TODO: deprecate
int csmngr_init_radio(struct csmngr_private *priv, struct radio_interface *radio,
		      enum csmngr_mode mode);

void csmngr_get_wifi_stations(struct csmngr_private *priv,
				struct radio_interface *iface);

int csmngr_register_misc_events(struct csmngr_private *priv);
int csmngr_unregister_misc_events(struct csmngr_private *priv);

int csmngr_register_nlevents(struct csmngr_private *priv);
void csmngr_unregister_nlevents(struct csmngr_private *priv);


uint32_t lookup_object(void *bus, const char *objname);

#ifdef HAS_UBUS
int csmngr_publish_object(struct csmngr_private *p, const char *objname);
int csmngr_remove_object(struct csmngr_private *p);

int csmngr_register_local_events(struct csmngr_private *p);
int csmngr_unregister_local_events(struct csmngr_private *p);
#else
static inline int csmngr_publish_object(struct csmngr_private *p, const char *objname)
{
	return 0;
}

static inline int csmngr_remove_object(struct csmngr_private *p)
{
	return 0;
}

static inline int csmngr_register_local_events(struct csmngr_private *p)
{
	return 0;
}

static inline int csmngr_unregister_local_events(struct csmngr_private *p)
{
	return 0;
}
#endif

char *csmode_str(enum csmngr_mode mode);
char *cs_state_str(enum csmngr_state s);
char *cs_criteria_str(enum cs_criteria_type s);
char *dfs_state_str(enum dfs_state state);
char *wifi_band_str(enum wifi_band band);

int csmngr_update_neigh_brport(struct csmngr_private *priv, char *brname);
char *csmngr_brport_to_ifname(struct csmngr_private *priv, uint16_t brport);
struct radio_interface *csmngr_ifname_to_interface(struct csmngr_private *priv, const char *ifname);

int csmngr_handle_ethport_carrier_off(struct csmngr_private *priv, const char *ifname);
int csmngr_handle_ethport_carrier_on(struct csmngr_private *priv, const char *ifname);

//int history_enqueue(void *priv, struct history_entry *he, uint32_t timeout);

int history_load_from_json_file(void *priv, const char *file);
int history_store_to_json_file(void *priv, const char *file);

int load_from_json_file(void *priv, const char *file);
int store_to_json_file(void *priv, const char *file);

#endif /* CSMNGR_H */
