/*
 * engine.c - WiFi channel selection manager s/m.
 *
 * Copyright (C) 2024 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdio.h>
#include <stdint.h>
#include <inttypes.h>
#include <stdlib.h>
#include <stdbool.h>
#include <float.h>
#include <math.h>

#include <easy/easy.h>
#include <wifi.h>

#include "util.h"
#include "debug.h"
#include "timer.h"
#include "csmngr.h"


////TODO: use from wifiutils
uint32_t wifi_bw_enum2MHz(enum wifi_bw bw)
{
	uint32_t bandwidth = 0;
	switch(bw) {
	case BW80:
		bandwidth = 80;
		break;
	case BW40:
		bandwidth = 40;
		break;
	case BW20:
		bandwidth = 20;
		break;
	case BW160:
	case BW8080:
		bandwidth = 160;
		break;
	case BW320:
		bandwidth = 320;
		break;
	default:
		break;
	}

	return bandwidth;
}

struct chanlist {
	int bw;
	enum wifi_chan_ext sb;
	int chans[24];
};

const int *chan2list_6g(int chan, int bw)
{
	static const struct chanlist chanlist[] = {
		{.bw = 20, .chans = {1, 0}},
		{.bw = 20, .chans = {5, 0}},
		{.bw = 20, .chans = {9, 0}},
		{.bw = 20, .chans = {13, 0}},
		{.bw = 20, .chans = {17, 0}},
		{.bw = 20, .chans = {21, 0}},
		{.bw = 20, .chans = {25, 0}},
		{.bw = 20, .chans = {29, 0}},
		{.bw = 20, .chans = {33, 0}},
		{.bw = 20, .chans = {37, 0}},
		{.bw = 20, .chans = {41, 0}},
		{.bw = 20, .chans = {45, 0}},
		{.bw = 20, .chans = {49, 0}},
		{.bw = 20, .chans = {53, 0}},
		{.bw = 20, .chans = {57, 0}},
		{.bw = 20, .chans = {61, 0}},
		{.bw = 20, .chans = {65, 0}},
		{.bw = 20, .chans = {69, 0}},
		{.bw = 20, .chans = {73, 0}},
		{.bw = 20, .chans = {77, 0}},
		{.bw = 20, .chans = {81, 0}},
		{.bw = 20, .chans = {85, 0}},
		{.bw = 20, .chans = {89, 0}},
		{.bw = 20, .chans = {93, 0}},
		/*TODO: bw = 20 UNII-6 ~ UNII-8 */
		{.bw = 40, .chans = {1, 5, 0}},
		{.bw = 40, .chans = {9, 13, 0}},
		{.bw = 40, .chans = {17, 21, 0}},
		{.bw = 40, .chans = {25, 29, 0}},
		{.bw = 40, .chans = {33, 37, 0}},
		{.bw = 40, .chans = {41, 45, 0}},
		{.bw = 40, .chans = {49, 53, 0}},
		{.bw = 40, .chans = {57, 61, 0}},
		{.bw = 40, .chans = {65, 69, 0}},
		{.bw = 40, .chans = {73, 77, 0}},
		{.bw = 40, .chans = {81, 85, 0}},
		{.bw = 40, .chans = {89, 93, 0}},
		/*TODO: bw = 40 UNII-6 ~ UNII-8 */
		{.bw = 80, .chans = {1, 5, 9, 13, 0}},
		{.bw = 80, .chans = {17, 21, 25, 29, 0}},
		{.bw = 80, .chans = {33, 37, 41, 45, 0}},
		{.bw = 80, .chans = {49, 53, 57, 61, 0}},
		{.bw = 80, .chans = {65, 69, 73, 77, 0}},
		{.bw = 80, .chans = {81, 85, 89, 93, 0}},
		/*TODO: bw = 80 UNII-6 ~ UNII-8 */
		{.bw = 160, .chans = {1, 5, 9, 13, 17, 21, 25, 29, 0}},
		{.bw = 160, .chans = {33, 37, 41, 45, 49, 53, 57, 61, 0}},
		{.bw = 160, .chans = {65, 69, 73, 77, 81, 85, 89, 93, 0}},
		/*TODO: bw = 160 UNII-6 ~ UNII-8 */
		{.bw = 320, .chans = {1, 5, 9, 13, 17, 21, 25, 29, 33, 37, 41, 45, 49, 53, 57, 61, 0}},
		{.bw = 320, .chans = {33, 37, 41, 45, 49, 53, 57, 61, 65, 69, 73, 77, 81, 85, 89, 93, 0}},
		/*TODO: bw = 320 UNII-6 ~ UNII-8 */
	};
	const int *chans;
	const int *start;
	int i;

	for (i = 0; i < ARRAY_SIZE(chanlist); i++) {
		if (chanlist[i].bw == bw) {
			chans = chanlist[i].chans;
			start = chans;
			while (*start) {
				if (*start == chan)
					return chans;
				start++;
			}
		}
	}

	return NULL;
}

const int *chan2list(int chan, int bw)
{
	static const struct chanlist chanlist[] = {
		{.bw = 20, .chans = {36, 0}},
		{.bw = 20, .chans = {40, 0}},
		{.bw = 20, .chans = {44, 0}},
		{.bw = 20, .chans = {48, 0}},
		{.bw = 20, .chans = {52, 0}},
		{.bw = 20, .chans = {56, 0}},
		{.bw = 20, .chans = {60, 0}},
		{.bw = 20, .chans = {64, 0}},
		{.bw = 20, .chans = {100, 0}},
		{.bw = 20, .chans = {104, 0}},
		{.bw = 20, .chans = {108, 0}},
		{.bw = 20, .chans = {112, 0}},
		{.bw = 20, .chans = {116, 0}},
		{.bw = 20, .chans = {120, 0}},
		{.bw = 20, .chans = {124, 0}},
		{.bw = 20, .chans = {128, 0}},
		{.bw = 20, .chans = {132, 0}},
		{.bw = 20, .chans = {136, 0}},
		{.bw = 20, .chans = {140, 0}},
		{.bw = 20, .chans = {144, 0}},
		{.bw = 20, .chans = {149, 0}},
		{.bw = 20, .chans = {153, 0}},
		{.bw = 20, .chans = {157, 0}},
		{.bw = 20, .chans = {161, 0}},
		{.bw = 20, .chans = {165, 0}},
		{.bw = 40, .chans = {36, 40, 0}},
		{.bw = 40, .chans = {44, 48, 0}},
		{.bw = 40, .chans = {52, 56, 0}},
		{.bw = 40, .chans = {60, 64, 0}},
		{.bw = 40, .chans = {100, 104, 0}},
		{.bw = 40, .chans = {108, 112, 0}},
		{.bw = 40, .chans = {116, 120, 0}},
		{.bw = 40, .chans = {124, 128, 0}},
		{.bw = 40, .chans = {132, 136, 0}},
		{.bw = 40, .chans = {140, 144, 0}},
		{.bw = 40, .chans = {149, 153, 0}},
		{.bw = 40, .chans = {157, 161, 0}},
		{.bw = 40, .chans = {165, 169, 0}},
		{.bw = 40, .chans = {173, 177, 0}},
		{.bw = 80, .chans = {36, 40, 44, 48, 0}},
		{.bw = 80, .chans = {52, 56, 60, 64, 0}},
		{.bw = 80, .chans = {100, 104, 108, 112, 0}},
		{.bw = 80, .chans = {116, 120, 124, 128, 0}},
		{.bw = 80, .chans = {132, 136, 140, 144, 0}},
		{.bw = 80, .chans = {149, 153, 157, 161, 0}},
		{.bw = 80, .chans = {165, 169, 173, 177, 0}},
		{.bw = 160, .chans = {36, 40, 44, 48, 52, 56, 60, 64, 0}},
		{.bw = 160, .chans = {100, 104, 108, 112, 116, 120, 124, 128, 0}},
		{.bw = 160, .chans = {149, 153, 157, 161, 165, 169, 173, 177, 0}},
	};
	const int *chans;
	const int *start;
	int i;

	for (i = 0; i < ARRAY_SIZE(chanlist); i++) {
		if (chanlist[i].bw == bw) {
			chans = chanlist[i].chans;
			start = chans;
			while (*start) {
				if (*start == chan)
					return chans;
				start++;
			}
		}
	}

	return NULL;
}

const int *chan2list_2g(int chan, int bw, enum wifi_chan_ext sideband)
{
	static const struct chanlist chanlist_2g[] = {
		{.bw = 20, .sb = EXTCH_NONE, .chans = {1, 0}},
		{.bw = 20, .sb = EXTCH_NONE, .chans = {6, 0}},
		{.bw = 20, .sb = EXTCH_NONE, .chans = {11, 0}},
		{.bw = 40, .sb = EXTCH_ABOVE, .chans = {1, 6, 0}},
		{.bw = 40, .sb = EXTCH_BELOW, .chans = {6, 1, 0}},
		{.bw = 40, .sb = EXTCH_ABOVE, .chans = {6, 11, 0}},
		{.bw = 40, .sb = EXTCH_BELOW, .chans = {11, 6, 0}},
	};
	int i;

	for (i = 0; i < ARRAY_SIZE(chanlist_2g); i++) {
		if (chanlist_2g[i].bw == bw &&
		    chanlist_2g[i].sb == sideband &&
		    chanlist_2g[i].chans[0] == chan) {
			return chanlist_2g[i].chans;
		}
	}

	return NULL;
}

const int *chan2list_2g_adj(int chan, int bw, enum wifi_chan_ext sideband)
{
	static const struct chanlist chanlist_2g_adj[] = {
		{.bw = 20, .chans = {1, 2, 3, 4, 0}},
		{.bw = 20, .chans = {2, 3, 4, 5, 1, 0}},
		{.bw = 20, .chans = {3, 4, 5, 6, 1, 2, 0}},
		{.bw = 20, .chans = {4, 5, 6, 7, 1, 2, 3, 0}},
		{.bw = 20, .chans = {5, 6, 7, 8, 2, 3, 4, 0}},
		{.bw = 20, .chans = {6, 7, 8, 9, 3, 4, 5, 0}},
		{.bw = 20, .chans = {7, 8, 9, 10, 4, 5, 6, 0}},
		{.bw = 20, .chans = {8, 9, 10, 11, 5, 6, 7, 0}},
		{.bw = 20, .chans = {9, 10, 11, 12, 6, 7, 8, 0}},
		{.bw = 20, .chans = {10, 11, 12, 13, 7, 8, 9, 0}},
		{.bw = 20, .chans = {11, 12, 13, 14, 8, 9, 10, 0}},
		{.bw = 20, .chans = {12, 13, 14, 9, 10, 11, 0}},
		{.bw = 20, .chans = {13, 14, 10, 11, 12, 0}},
		{.bw = 20, .chans = {14, 11, 12, 13, 0}},

		{.bw = 40, .sb = EXTCH_ABOVE, .chans = {1, 2, 3, 4, 5, 6, 0}},
		{.bw = 40, .sb = EXTCH_ABOVE, .chans = {2, 3, 4, 5, 6, 7, 1, 0}},
		{.bw = 40, .sb = EXTCH_ABOVE, .chans = {3, 4, 5, 6, 7, 8, 1, 2, 0}},
		{.bw = 40, .sb = EXTCH_ABOVE, .chans = {4, 5, 6, 7, 8, 9, 1, 2, 3, 0}},

		{.bw = 40, .sb = EXTCH_ABOVE, .chans = {5, 6, 7, 8,  2, 3, 4, 9, 10, 11, 0}},
		{.bw = 40, .sb = EXTCH_ABOVE, .chans = {6, 7, 8, 9, 3, 4, 5, 10, 11, 12, 0}},
		{.bw = 40, .sb = EXTCH_ABOVE, .chans = {7, 8, 9, 10, 4, 5, 6, 11, 12, 13, 0}},
		{.bw = 40, .sb = EXTCH_ABOVE, .chans = {8, 9, 10, 11, 5, 6, 7, 12, 13, 14, 0}},
		{.bw = 40, .sb = EXTCH_ABOVE, .chans = {9, 10, 11, 12, 6, 7, 8, 13, 14, 0}},

		{.bw = 40, .sb = EXTCH_BELOW, .chans = {5, 6, 7, 8, 1, 2, 3, 4, 0}},
		{.bw = 40, .sb = EXTCH_BELOW, .chans = {6, 7, 8, 9, 1, 2, 3, 4, 5, 0}},
		{.bw = 40, .sb = EXTCH_BELOW, .chans = {7, 8, 9, 10, 1, 2, 3, 4, 5, 6, 0}},
		{.bw = 40, .sb = EXTCH_BELOW, .chans = {8, 9, 10, 11, 2, 3, 4, 5, 6, 7, 0}},
		{.bw = 40, .sb = EXTCH_BELOW, .chans = {9, 10, 11, 12, 3, 4, 5, 6, 7, 8, 0}},

		{.bw = 40, .sb = EXTCH_BELOW, .chans = {10, 11, 12, 13, 14, 5, 6, 7, 8, 9, 0}},
		{.bw = 40, .sb = EXTCH_BELOW, .chans = {11, 12, 13, 14, 6, 7, 8, 9, 10, 0}},
		{.bw = 40, .sb = EXTCH_BELOW, .chans = {12, 13, 14, 7, 8, 9, 10, 11, 0}},
		{.bw = 40, .sb = EXTCH_BELOW, .chans = {13, 14, 8, 9, 10, 11, 12, 0}},
		{.bw = 40, .sb = EXTCH_BELOW, .chans = {14, 9, 10, 11, 12, 13, 0}},
	};
	int i;

	for (i = 0; i < ARRAY_SIZE(chanlist_2g_adj); i++) {
		if (chanlist_2g_adj[i].bw == bw &&
		    chanlist_2g_adj[i].sb == sideband &&
		    chanlist_2g_adj[i].chans[0] == chan) {
			return chanlist_2g_adj[i].chans;
		}
	}

	return NULL;
}

static const int *radio_adjacent_channels(enum wifi_band band, int chan, int bw,
					  enum wifi_chan_ext sideband)
{
	if (band == BAND_6)
		return chan2list_6g(chan, bw);
	else if (band == BAND_5)
		return chan2list(chan, bw);

	return chan2list_2g_adj(chan, bw, sideband);
}

static const int *radio_chan2list(enum wifi_band band, int chan, int bw,
				  enum wifi_chan_ext sideband)
{
	if (band == BAND_6)
		return chan2list_6g(chan, bw);
	else if (band == BAND_5)
		return chan2list(chan, bw);

	return chan2list_2g(chan, bw, sideband);
}

static bool radio_opclass_dfs_required(int channel, enum wifi_bw bw,
				       struct chan_entry *chan, int chan_num)
{
	const int *chans;
	int j;

	chans = chan2list(channel, wifi_bw_enum2MHz(bw));

	while (chans && *chans) {
		for (j = 0; j < chan_num; j++) {
			if (*chans == chan[j].channel) {
				if (chan[j].dfs)
					return true;
			}
		}
		chans++;
	}

	return false;
}

static bool radio_opclass_dfs_unavailable(int channel, enum wifi_bw bw,
					  struct chan_entry *chan, int chan_num)
{
	const int *chans;
	int j;

	chans = chan2list(channel, wifi_bw_enum2MHz(bw));

	/* Check any channel in NOP state (unavailable) */
	while (chans && *chans) {
		for (j = 0; j < chan_num; j++) {
			if (*chans == chan[j].channel) {
				if (!chan[j].dfs)
					continue;
				if (chan[j].dfs_state == WIFI_DFS_STATE_UNAVAILABLE)
					return true;
			}
		}
		chans++;
	}

	return false;
}

static bool radio_opclass_dfs_available(int channel, enum wifi_bw bw,
					struct chan_entry *chan, int chan_num)
{
	const int *chans;
	int j;

	chans = chan2list(channel, wifi_bw_enum2MHz(bw));

	/* Check all DFS channels pass CAC (available) */
	while (chans && *chans) {
		for (j = 0; j < chan_num; j++) {
			if (*chans == chan[j].channel) {
				if (!chan[j].dfs)
					continue;
				if (chan[j].dfs_state != WIFI_DFS_STATE_AVAILABLE)
					return false;
			}
		}
		chans++;
	}

	return true;
}

static bool radio_opclass_dfs_cac(int channel, enum wifi_bw bw,
				  struct chan_entry *chan, int chan_num)
{
	const int *chans;
	int j;

	chans = chan2list(channel, wifi_bw_enum2MHz(bw));

	/* Check all DFS channels CAC ongoing (cac) */
	while (chans && *chans) {
		for (j = 0; j < chan_num; j++) {
			if (*chans == chan[j].channel) {
				if (!chan[j].dfs)
					continue;
				if (chan[j].dfs_state != WIFI_DFS_STATE_CAC)
					return false;
			}
		}
		chans++;
	}

	return true;
}

static bool radio_opclass_dfs_usable(int channel, enum wifi_bw bw,
				     struct chan_entry *chan, int chan_num)
{
	const int *chans;
	int j;

	chans = chan2list(channel, wifi_bw_enum2MHz(bw));

	/* Check all DFS channels require CAC (usable) */
	while (chans && *chans) {
		for (j = 0; j < chan_num; j++) {
			if (*chans == chan[j].channel) {
				if (!chan[j].dfs)
					continue;
				if (chan[j].dfs_state != WIFI_DFS_STATE_USABLE)
					return false;
			}
		}
		chans++;
	}

	return true;
}

static enum dfs_state radio_opclass_dfs_state(int channel,
					      enum wifi_bw bw,
					      struct chan_entry *chan,
					      int chan_num)
{
	if (radio_opclass_dfs_unavailable(channel, bw, chan, chan_num))
		return WIFI_DFS_STATE_UNAVAILABLE;

	if (radio_opclass_dfs_available(channel, bw, chan, chan_num))
		return WIFI_DFS_STATE_AVAILABLE;

	if (radio_opclass_dfs_cac(channel, bw, chan, chan_num))
		return WIFI_DFS_STATE_CAC;

	if (radio_opclass_dfs_usable(channel, bw, chan, chan_num))
		return WIFI_DFS_STATE_USABLE;

	return WIFI_DFS_STATE_USABLE;
}

static uint32_t radio_dfs_cac_time(int channel,
				   enum wifi_bw bw,
				   struct chan_entry *chan,
				   int chan_num)
{
	uint32_t cac_time = 0;
	const int *channels;
	int bandwidth;
	int i;

	/* Get adjacent channels we should care */
	bandwidth = wifi_bw_enum2MHz(bw);
	channels = chan2list(channel, bandwidth);

	/* Get worst (highest) CAC value */
	while (channels && *channels) {
		for (i = 0; i < chan_num; i++) {
			if (chan[i].channel != *channels)
				continue;

			if (chan[i].cac_time > cac_time) {
				cac_time = chan[i].cac_time;
				//dbg("[%s] cac_time chan %d cac_time %d new cac_time %d\n",
				//    name, *channels, chan[i].cac_time, cac_time);
			}
			break;
		}
		channels++;
	}

	return cac_time;
}

static uint32_t radio_dfs_nop_time(int channel,
				   enum wifi_bw bw,
				   struct chan_entry *chan,
				   int chan_num)
{
	uint32_t nop_time = 0;
	const int *channels;
	int bandwidth;
	int i;

	/* Get adjacent channels we should care */
	bandwidth = wifi_bw_enum2MHz(bw);
	channels = chan2list(channel, bandwidth);

	/* Get worst (highest) NOP value */
	while (channels && *channels) {
		for (i = 0; i < chan_num; i++) {
			if (chan[i].channel != *channels)
				continue;

			if (chan[i].nop_time > nop_time) {
				nop_time = chan[i].nop_time;
				// dbg("nop_time chan %d noptime %d new nop_time %d\n",
				//     *channels, chan[i].nop_time, nop_time);
			}
			break;
		}
		channels++;
	}

	return nop_time;
}

static int radio_opclass_recalc_noise(struct radio_interface *r,
				      int channel,
				      enum wifi_bw bw,
				      enum wifi_chan_ext sideband,
				      struct chan_entry *chan,
				      int chan_num)
{
	int noise = -255;
	const int *channels;
	int bandwidth;
	int i;

	/* Get adjacent channels we should care */
	bandwidth = wifi_bw_enum2MHz(bw);
	channels = radio_chan2list(r->band, channel, bandwidth, sideband);

	/* Get worst (highest) noise value */
	while (channels && *channels) {
		for (i = 0; i < chan_num; i++) {
			if (chan[i].channel != *channels)
				continue;

			if (chan[i].noise > noise)
				noise = chan[i].noise;
			break;
		}
		channels++;
	}

	dbg("[%s] channel %d/%d updated noise %d\n", r->ifname, channel, bandwidth, noise);
	return noise;
}

#if 0
/* return average busyness for this channel/bandwidth grp */
static uint8_t radio_opclass_recalc_busy(struct radio_interface *r,
					 int channel,
					 enum wifi_bw bw,
					 enum wifi_chan_ext sideband,
					 struct chan_entry *chan,
					 int chan_num)
{
	const int *channels;
	int bandwidth;
	int busy = 0;
	int busy_num = 0;
	int i;

	/* Get adjacent channels we should care */
	bandwidth = wifi_bw_enum2MHz(bw);
	channels = radio_chan2list(r->band, channel, bandwidth, sideband);

	/* Get average value */
	while (channels && *channels) {
		for (i = 0; i < chan_num; i++) {
			if (chan[i].channel != *channels)
				continue;

			/* If invalid value for one channel - mark bw invalid */
			if (chan[i].busy > 100)
				return 255;

			busy += chan[i].busy;
			busy_num++;
			break;
		}
		channels++;
	}

	if (!busy_num)
		return 255;

	busy = busy/busy_num;
	dbg("[%s] avg busy %d\n", r->ifname, busy);
	return (uint8_t)((busy > 255) ? 255 : busy);
}
#endif

/* return average interference-factor for this channel/bandwidth grp */
static double radio_opclass_recalc_intfx(struct radio_interface *r,
					      int channel,
					      enum wifi_bw bw,
					      enum wifi_chan_ext sideband,
					      struct chan_entry *chan,
					      int chan_num)
{
	double intfx = 0;
	double num_intfx = 0;
	const int *channels;
	int bandwidth;
	int i;


	/* Get adjacent channels we should care */
	bandwidth = wifi_bw_enum2MHz(bw);
	channels = radio_chan2list(r->band, channel, bandwidth, sideband);
	dbg("%s: channel = %u, bw = %u\n", __func__, channel, bandwidth);

	/* Get average value */
	while (channels && *channels) {
		for (i = 0; i < chan_num; i++) {
			int chidx = -1;

			if (chan[i].channel != *channels)
				continue;

			chidx = radio_get_chinfo_index_by_channel(r, *channels);
			if (chidx == -1) {
				warn("%s: channel %u not in surveylist of %s\n", __func__, *channels, r->ifname);
				continue;
			}

			intfx += r->intfx[chidx];
			num_intfx += 1;
			break;
		}
		channels++;
	}

	if (!num_intfx)
		return (double)-1;

	intfx /= num_intfx;

	dbg("%s: %u/%u avg interference-factor = %g\n", r->ifname, channel, bandwidth, intfx);
	return intfx;
}

/* return total number of scanresults for this channel/bandwidth */
static uint8_t radio_opclass_bss_num(struct radio_interface *r,
				     int channel, enum wifi_bw bw,
				     enum wifi_chan_ext sideband, struct chan_entry *chan,
				     int chan_num, struct wifi_bss *bss, int bss_num)
{
	const int *chans, *channels;
	int bandwidth;
	uint8_t num = 0;
	int i;

	bandwidth = wifi_bw_enum2MHz(bw);
	channels = radio_adjacent_channels(r->band, channel, bandwidth, sideband);
	for (i = 0; i < bss_num; i++) {
		chans = channels;
		while (chans && *chans) {
			if (bss[i].channel == *chans) {
				num++;
				//dbg("[%s] chan %d/%d new bss (bss %s chan %d) num: %d\n",
				//    name, channel, bandwidth, bss[i].ssid, *chans, num);
			}
			chans++;
		}
	}

	return num;
}

static bool in_preferred_channels(struct radio_interface *r,
				  uint8_t *preflist, int num_pref,
				  int channel, enum wifi_bw bw,
				  enum wifi_chan_ext sideband)
{
#ifdef PREFLIST_HAS_PRIMARY_CHANNELS_ONLY
	const int *channels;
	int bandwidth;

	bandwidth = wifi_bw_enum2MHz(bw);
	channels = radio_chan2list(r->band, channel, bandwidth, sideband);

	while (channels && *channels) {
		bool found = false;

		for (int i = 0; i < num_pref; i++) {
			if (preflist[i] == *channels) {
				found = true;
				break;
			}
		}

		if (!found) {
			dbg("%s: channel = %u NOT in preferred list\n", __func__, *channels);
			return false;
		}
		channels++;
	}

	return true;
#else
		for (int i = 0; i < num_pref; i++) {
			if (preflist[i] == channel) {
				return true;
			}
		}

		return false;

#endif /* PREFLIST_HAS_PRIMARY_CHANNELS_ONLY */
}

static bool is_6ghz_psc_channel(int channel)
{
	uint8_t psc_channels[] = { 5, 21, 37, 53, 69, 85, 101, 117,
				   133, 149, 165, 181, 197, 213, 229 };

	for (int i = 0; i < ARRAY_SIZE(psc_channels); i++) {
		if (psc_channels[i] == channel)
			return true;
	}

	return false;
}

int radio_get_chinfo_index_by_channel(struct radio_interface *r, uint32_t channel)
{
	int i;

	for (i = 0; i < r->num_chinfo; i++) {
		struct chan_entry *e = &r->chinfo[i];

		if (e->channel == channel)
			return i;
	}

	return -1;
}

static int bw2idx(enum wifi_bw bw)
{
	int idx = -1;

	switch (bw) {
	case BW20 ... BW160:
		idx = bw;
		break;
	case BW320:
		idx = 4;
	default:
		break;
	}

	return idx;
}

int radio_calc_score(struct radio_interface *r)
{
	int idx = 0;

	for (idx = 0; idx < NUM_BANDWIDTHS; idx++) {
		for (int k = 0; k < r->verdict[idx].nent; k++) {
			double score = 0;

			for (int i = 0; i < NUM_CS_CRITERIA; i++) {
				switch (i) {
				case INTFX_FACTOR:
					{
						double val = 0;

						memcpy(&val, r->verdict[idx].ent[k].opt[i].wtdata, sizeof(val));
						score += val;
					}
					break;
				case NUM_OBSS:
				case REG_TXPOWER:
					{
						uint32_t val = buf_get_be32(r->verdict[idx].ent[k].opt[i].wtdata);

						score += (double)val;
					}
					break;
				default:
					break;
				}
			}

			r->verdict[idx].ent[k].score = score;
			dbg("%s: channel = %3u, bw = %u, SCORE = %g\n",
			    r->ifname,
			    r->verdict[idx].ent[k].channel,
			    r->verdict[idx].ent[k].bw,
			    score);
		}
	}

	/* update verdict for each bandwidth */
	for (idx = 0; idx < NUM_BANDWIDTHS; idx++) {
		double min_score = DBL_MAX;
		int k = 0;

		for (int i = 0; i < r->verdict[idx].nent; i++) {
			if (r->verdict[idx].ent[i].score > 0 && min_score > r->verdict[idx].ent[i].score) {
				min_score = r->verdict[idx].ent[i].score;
				k = i;
			}
		}

		if (min_score == DBL_MAX)
			continue;

		r->verdict[idx].opclass = r->verdict[idx].ent[k].opclass;
		r->verdict[idx].channel = r->verdict[idx].ent[k].channel;
		r->verdict[idx].bw = r->verdict[idx].ent[k].bw;
		r->verdict[idx].score = r->verdict[idx].ent[k].score;
		dbg("%s: BEST (%3u MHz): Opclass: %u  ctrl-channel = %u  score = %g\n",
		    r->ifname, wifi_bw_enum2MHz(r->verdict[idx].bw),
		    r->verdict[idx].opclass, r->verdict[idx].channel,
		    r->verdict[idx].score);
	}

	return 0;
}

int radio_update_cs_criteria_score(struct radio_interface *r,
				   enum cs_criteria_type c,
				   uint32_t opclass,
				   uint32_t channel,
				   enum wifi_bw bw,
				   void *value)
{
	int idx = bw2idx(bw);
	int i;

	if (idx == -1) {
		dbg("%s: %s - unsupported ch = %u, bw %u\n", __func__,
		    r->ifname, channel, bw);
		return 0;
	}

	for (i = 0; i < r->verdict[idx].nent; i++) {
		if (r->verdict[idx].ent[i].channel == channel)
			break;
	}

	switch (c) {
	case INTFX_FACTOR:
		{
			double intfx = 0;
			double wintfx = 0;

			memcpy(&intfx, value, sizeof(intfx));
			wintfx = (r->cs_attr_weight[INTFX_FACTOR] * intfx) / (double)100;

			if (i == r->verdict[idx].nent)
				r->verdict[idx].nent++;

			if (r->verdict[idx].nent > 64)	//TODO:
				return 0;

			r->verdict[idx].ent[i].opclass = opclass;
			r->verdict[idx].ent[i].channel = channel;
			r->verdict[idx].ent[i].bw = bw;
			memcpy(r->verdict[idx].ent[i].opt[INTFX_FACTOR].data, &intfx, sizeof(intfx));
			memcpy(r->verdict[idx].ent[i].opt[INTFX_FACTOR].wtdata, &wintfx, sizeof(wintfx));
		}
		break;
	case NUM_OBSS:
		{
			uint32_t num_obss = 0;
			uint32_t wnum_obss = 0;

			memcpy(&num_obss, value, sizeof(num_obss));
			wnum_obss = (r->cs_attr_weight[NUM_OBSS] * num_obss ) / 100;

			if (i == r->verdict[idx].nent)
				r->verdict[idx].nent++;

			if (r->verdict[idx].nent > 64)	//TODO:
				return 0;

			r->verdict[idx].ent[i].opclass = opclass;
			r->verdict[idx].ent[i].channel = channel;
			r->verdict[idx].ent[i].bw = bw;
			buf_put_be32(r->verdict[idx].ent[i].opt[NUM_OBSS].data, num_obss);
			buf_put_be32(r->verdict[idx].ent[i].opt[NUM_OBSS].wtdata, wnum_obss);
		}
		break;
	case REG_TXPOWER:
		/* TODO */
		break;
	default:
		return 0;
	}

	return 0;
}

int csmngr_find_best_channel(struct radio_interface *r)
{
	for (int i = 0; i < r->num_supp_opclass; i++) {
		struct wifi_opclass *op = &r->supp_opclass[i];


		/* skip opclass not supported by radio */
		if (op->band != r->band)
			continue;

		for (int j = 0; j < op->opchannel.num; j++) {
			struct chan_entry *ch = &op->opchannel.ch[j];
			uint8_t channel = ch->channel;
			enum wifi_bw bw = op->bw;
			enum wifi_chan_ext sideband;
			struct channel_data *chdata = NULL;
			double intfx = 0;


			/* skip opclass if bw is not supported */
			if (!(BIT(op->bw) & r->supp_bandwidths))
				continue;

			switch (op->g_opclass) {
			case 83:
				sideband = EXTCH_ABOVE;
				break;
			case 84:
				sideband = EXTCH_BELOW;
				break;
			default:
				sideband = EXTCH_NONE;
				break;
			}

			//TODO: exclude/limit bandwidth through config
			if (r->band == BAND_2 && bw > BW20) {
				dbg("%s: skip 40MHz bandwidth opclass in 2.4GHz\n", r->ifname);
				continue;
			}

			/* In 5GHz, exclude DFS channels if requested */
			if (r->band == BAND_5 && r->exclude_dfs && ch->dfs)
				continue;

			/* In 6GHz, exclude non-PSC by default, unless allowed */
			if (r->band == BAND_6 && r->exclude_non_psc && !is_6ghz_psc_channel(channel))
				continue;

			//TODO: channel bias from config

			/* prefer 1, 6 or 11 if configured such */
			if (r->band == BAND_2 && r->radio2.pick_1_6_11 &&
			    !(channel == 1 || channel == 6 || channel == 11)) {
				dbg("%s: prefer channels 1,6 or 11; skip %u\n", r->ifname, channel);
				continue;
			}

			/* BW8080 mean upper 80MHz (80+) */
			if (op->bw == BW8080)
				bw = BW80;


			chdata = radio_lookup_channel_data(r, ch->channel);
			if (!chdata)
				continue;

			/* skip if not in preferred channels */
			if (r->num_pref_channels &&
			    !(in_preferred_channels(r,
						    r->pref_channels,
						    r->num_pref_channels,
						    channel, bw, sideband))) {

				dbg("%s: skip non-preferred channel %u bw %u\n",
				    r->ifname, channel, wifi_bw_enum2MHz(bw));
				continue;
			}

			/* Check if dfs required for channel/bw */
			ch->dfs = radio_opclass_dfs_required(ch->channel,
							     bw,
							     r->chinfo,
							     r->num_chinfo);


			ch->dfs_state = radio_opclass_dfs_state(ch->channel,
								bw,
								r->chinfo,
								r->num_chinfo);
			if (ch->dfs_state == WIFI_DFS_STATE_USABLE) {
				dbg("%s: skip DFS not-cleared channel %u bw %u\n",
				    r->ifname, channel, wifi_bw_enum2MHz(bw));
				continue;
			}

			ch->noise = radio_opclass_recalc_noise(r,
							       ch->channel,
							       bw,
							       sideband,
							       r->chinfo,
							       r->num_chinfo);

#if 0
			ch->busy = radio_opclass_recalc_busy(r,
							     ch->channel,
							     bw,
							     sideband,
							     r->chinfo,
							     r->num_chinfo);
#endif

			intfx = radio_opclass_recalc_intfx(r,
							   ch->channel,
							   bw,
							   sideband,
							   r->chinfo,
							   r->num_chinfo);

			/* update score interference-factor */
			if (intfx >= 0) {
				radio_update_cs_criteria_score(r,
							       INTFX_FACTOR,
							       op->opclass,
							       ch->channel,
							       bw,
							       &intfx);
			}

			ch->bss_num = radio_opclass_bss_num(r,
							    ch->channel,
							    bw,
							    sideband,
							    r->chinfo,
							    r->num_chinfo,
							    chdata->scanresults,
							    chdata->num_scanresults);

			/* update score for num-obss */
			if (ch->bss_num > 0) {
				radio_update_cs_criteria_score(r,
							       NUM_OBSS,
							       op->opclass,
							       ch->channel,
							       bw,
							       &ch->bss_num);
			}


			ch->cac_time = radio_dfs_cac_time(ch->channel,
							  bw,
							  r->chinfo,
							  r->num_chinfo);

			ch->nop_time = radio_dfs_nop_time(ch->channel,
							  bw,
							  r->chinfo,
							  r->num_chinfo);


		}
	}

	radio_calc_score(r);
	return 0;
}

int csmngr_calculate_interference_factor(struct radio_interface *r)
{
	double lowest_intfactor = 0;
	uint32_t candiate_channel = 0;

	for (int i = 0; i < r->num_chinfo; i++) {
		struct chan_entry *e = &r->chinfo[i];
		double factor, busy, total;


		if (e->survey.cca_time == 0) {
			dbg("Channel %d (%u MHz) missing survey data\n", e->channel, e->freq);
			continue;
		}

		total = e->survey.cca_time;
		busy = e->survey.channel_busy;
		if (busy == 0 && e->survey.rx_airtime > 0)
			busy = e->survey.rx_airtime;

		if (e->survey.tx_airtime > 0) {
			busy -= e->survey.tx_airtime;
			total -= e->survey.tx_airtime;
		}

		factor = pow(10, e->noise / 5.0L) +
				(total ? (busy / total) : 0) *
					pow(2, pow(10, (double)e->noise / 10.0L) -
					pow(10, (double)r->nf / 10.0L));

		if (i == 0) {
			lowest_intfactor = factor;
			candiate_channel = e->channel;
		} else {
			if (factor < lowest_intfactor) {
				lowest_intfactor = factor;
				candiate_channel = e->channel;
			}
		}

		r->intfx[i] = factor;
		dbg("%s: Ch = %3u interference-factor = %g\n", __func__, e->channel, factor);
	}

	dbg("Channel %u has lowest interference-factor %g\n", candiate_channel, lowest_intfactor);
	return 0;
}

int wifi_channel2freq(enum wifi_band band, uint32_t channel)
{
	if (channel == 0)
		return -EINVAL;

	switch (band) {
	case BAND_2:
		if (channel <= 14)
			return (channel < 14) ? 2407 + channel * 5 : 2484;
		break;
	case BAND_5:
		if (channel >= 36 && channel <= 196)
			return (channel >= 182) ? 4000 + channel * 5 : 5000 + channel * 5;
		break;
	case BAND_6:
		if (channel >= 1 && channel <= 233)
			return 5950 + channel * 5;
		break;
	default:
		break;
	}

	return -EINVAL;
}

int csmngr_channel_switch(struct radio_interface *r, uint32_t opclass,
			  uint32_t channel, uint32_t bandwidth)
{
	struct chan_switch_param param = {
					.count = 5,
					.blocktx = 1,
					.bandwidth = bandwidth,
				};
	int sco = opclass == 83 ? 1 : opclass == 84 ? -1 : 0;
	char *ap_ifname = NULL;

	param.freq = wifi_channel2freq(r->band, channel);
	param.sec_chan_offset = sco;
	param.eht = !!(r->info.oper_std & WIFI_BE) ? true : false;
	param.he = !!(r->info.oper_std & WIFI_AX) ? true : false;
	param.vht = !!(r->info.oper_std & WIFI_AC) ? true : false;
	param.ht = !!(r->info.oper_std & WIFI_N) ? true : false;

	ap_ifname = csmngr_radio_get_first_ap_ifname(r);
	if (!ap_ifname) {
		dbg("%s: no AP mode interface for channel switch\n", r->ifname);
		return -1;
	}

	dbg("wifi_chan_switch(%s): freq = %u, sco = %u, eht = %u, he = %u, vht = %u, ht = %u\n",
	    ap_ifname, param.freq, sco, param.eht, param.he, param.vht, param.ht);

	return wifi_chan_switch(ap_ifname, &param);
}

int csmngr_process_survey_data(struct radio_interface *r)
{
	enum wifi_bw bw = BW_UNKNOWN;
	uint32_t opclass = 0;
	uint32_t channel = 0;
	int ret = 0;

	dbg("%s: '%s'\n", __func__, r->ifname);
	ret = csmngr_calculate_interference_factor(r);
	//TODO: handle error? maybe mark channels not having chinfo data invalid
	//if (ret) {
	//	return ret;
	//}

	csmngr_find_best_channel(r);
	r->refresh = 0;

	/* compare with current channel/bw */
	for (int i = NUM_BANDWIDTHS - 1; i >= 0; i--) {
		if (r->verdict[i].opclass == 0)
			continue;

		if (wifi_bw_enum2MHz(r->info.curr_bw) < wifi_bw_enum2MHz(r->verdict[i].bw)) {
			opclass = r->verdict[i].opclass;
			channel = r->verdict[i].channel;
			bw = r->verdict[i].bw;

			dbg("%s: Better %u (%3u MHz) in opclass: %u\n",
			    r->ifname, channel, wifi_bw_enum2MHz(bw), opclass);
			break;
		} else if (wifi_bw_enum2MHz(r->info.curr_bw) == wifi_bw_enum2MHz(r->verdict[i].bw)) {
			if (r->info.channel != r->verdict[i].channel) {
				opclass = r->verdict[i].opclass;
				channel = r->verdict[i].channel;
				bw = r->verdict[i].bw;

				dbg("%s: Better %u (%3u MHz) in opclass: %u\n",
				    r->ifname, channel, wifi_bw_enum2MHz(bw), opclass);
				break;
			}
		}
	}

	if (!channel || (channel == r->info.channel && bw == r->info.curr_bw)) {
		info("%s: current channel %u (%3u MHz) is OPTIMAL\n", r->ifname,
		     r->info.channel, wifi_bw_enum2MHz(r->info.curr_bw));

		return 0;
	}

	//TODO: check other conditions allow setting channel or not

	info("%s: Set OPTIMAL channel %u (%3u MHz) in opclass %u\n",
	     r->ifname, channel, wifi_bw_enum2MHz(bw), opclass);

	ret = csmngr_channel_switch(r, opclass, channel, wifi_bw_enum2MHz(bw));
	if (ret) {
		info("%s - failed to set channel %u/%u\n", r->ifname,
		     channel, wifi_bw_enum2MHz(bw));
	}

	return 0;
}

int csmngr_radio_scan(struct radio_interface *r, uint32_t channel)
{
	struct scan_param_ex sp = {0};
	int ret = 0;

	sp.num_freq = 1;
	sp.freq[0] = wifi_channel_to_freq_ex(channel, r->band);

	dbg("%s: Trigger scan (channel = %u, freq = %u)\n", __func__, channel, sp.freq[0]);

	r->scan_channel = channel;

	if (r->multiband)
		ret = wifi_scan_band_ex(r->phyname, r->band, &sp);
	else
		ret = wifi_scan_ex(r->phyname, &sp);

	return ret;
}

bool csmngr_radio_channel_is_unavailable(struct radio_interface *r, uint32_t channel)
{
	if (r->band != BAND_5)
		return false;

	for (int i = 0; i < r->num_chinfo; i++) {
		struct chan_entry *e = &r->chinfo[i];

		if (e->channel == channel && e->dfs &&
		    e->dfs_state == WIFI_DFS_STATE_UNAVAILABLE) {
			return true;
		}
	}

	return false;
}

bool csmngr_radio_channel_need_cac(struct radio_interface *r, uint32_t channel)
{
	if (r->band != BAND_5)
		return false;

	for (int i = 0; i < r->num_chinfo; i++) {
		struct chan_entry *e = &r->chinfo[i];

		if (e->channel == channel && e->dfs &&
		    e->dfs_state == WIFI_DFS_STATE_USABLE) {
			return true;
		}
	}

	return false;
}

int csmngr_radio_next_cac_channel_default(struct radio_interface *r,
					  uint32_t *cac_channel,
					  enum wifi_bw *cac_bw)
{
	struct chan_entry *e = NULL, *en = NULL;
	int idx = 0;
	int i;

#define MAX_BANDWIDTH_CLEAR_ONESHOT	80	/* or 160, 40, 20 */

	/* With oneshot clear set to 80MHz, the channels preclear order is -
	 * 52-64, 100-112, 116-128, 132-144
	 */

	*cac_channel = 0;
	*cac_bw = BW_UNKNOWN;

	for (i = 0; i < r->num_chinfo; i++) {
		e = &r->chinfo[i];

		if (e->dfs && e->dfs_state == WIFI_DFS_STATE_USABLE) {
			idx = i;	/* next channel for CAC'ng */
			break;
		}
	}

	if (idx == r->num_chinfo)
		return 0;

	r->ncac = idx;

	for (i = idx; i < r->num_chinfo; i++) {
		int bandwidth = 20;
		int m = 1;

		e = &r->chinfo[i];
		m = 1;
		dbg("start = %u\n", e->channel);

		while (i + 1 < r->num_chinfo &&
		       bandwidth < MAX_BANDWIDTH_CLEAR_ONESHOT &&
		       (en = &r->chinfo[i + 1]) &&
		       en->dfs && en->dfs_state == WIFI_DFS_STATE_USABLE &&
		       (en->channel - e->channel == 4 * m)
		       /* && en->cac_time == e->cac_time */) {

			bandwidth += 20;
			dbg("[%d]: ( %d ~ %d ) bandwidth = %d\n",
			    i, e->channel, en->channel, bandwidth);

			i++;
			m++;
		};

		if (e->dfs) {
			enum wifi_bw bw = BW20;

			if (bandwidth == 160)
				bw = BW160;
			else if (bandwidth == 80)
				bw = BW80;
			else if (bandwidth == 40)
				bw = BW40;

			if (bw == BW20)
				bandwidth = 20;

			dbg("CAC: start-ch = %u (bandwidth '%d')\n",
			    e->channel, bandwidth);

			*cac_channel = e->channel;
			*cac_bw = bw;
			break;
		}
	}

	return 0;
}

int csmngr_radio_next_cac_channel(struct radio_interface *r,
				  uint32_t *channel, enum wifi_bw *bw)
{
	if (r->band != BAND_5 || r->cac_ongoing)
		return -1;

	*channel = 0;
	*bw = BW_UNKNOWN;

	if (r->radio5.num_bgcac == 0)
		return csmngr_radio_next_cac_channel_default(r, channel, bw);

	if (r->radio5.next_cac < r->radio5.num_bgcac) {
		*channel = r->radio5.caclist[r->radio5.next_cac].channel;
		*bw = r->radio5.caclist[r->radio5.next_cac].bw;
		r->radio5.next_cac++;
		return 0;
	}

	return 0;
}

bool csmngr_radio_should_cac(struct radio_interface *r)
{
	if (r->band != BAND_5 || r->cac_ongoing)
		return false;

	for (int i = 0; i < r->num_chinfo; i++) {
		struct chan_entry *e = &r->chinfo[i];

		if (e->dfs && e->dfs_state == WIFI_DFS_STATE_USABLE) {
			r->ncac = i;	/* next channel for CAC'ng */
			return true;
		}
	}

	return false;
}

int csmngr_radio_trigger_cac(struct radio_interface *r, uint32_t channel, enum wifi_bw bw)
{
	dbg("CAC: start-ch = %u (bw = %u)\n", channel, bw);
	return wifi_start_cac(r->phyname, channel, bw, WIFI_CAC_MIMO_REDUCED);
}

bool csmngr_radio_is_idle(struct radio_interface *r)
{
	double busy_pct =
		((double)r->info.diag.channel_busy /
		 (double)r->info.diag.cca_time) * 100.0;

	double txrx_pct =
		((double)(r->info.diag.tx_airtime + r->info.diag.rx_airtime) /
		 (double)r->info.diag.cca_time) * 100.0;

	dbg("%s: busy = %g %%, txrx = %g %%\n", r->ifname, busy_pct, txrx_pct);

	return (txrx_pct < (double)r->idle_threshold) ? true : false;
}

int csmngr_radio_run(struct radio_interface *r)
{
	int ret = 0;

	switch (r->state) {
	case CS_STATE_IDLE:
		/* update on-channel surveyinfo */
		csmngr_radio_collect_channels_info(r);

		if (r->refresh && csmngr_radio_is_idle(r)) {
			dbg("%s: is below idle-threshold. Triggering scan ...\n", r->ifname);
			r->state = CS_STATE_SCANNING;
			timer_set(&r->runtimer, 200);
			break;
		}

		if (csmngr_radio_should_cac(r)) {
			r->state = CS_STATE_CAC;
			timer_set(&r->runtimer, 200);
			break;
		}

		break;
	case CS_STATE_SCANNING:
		while (r->nscan < r->num_supp_channels) {
			if (csmngr_radio_channel_need_cac(r, r->supp_channels[r->nscan])) {
				dbg("%s: Skip %u as it requires CAC before can be used\n",
				    r->ifname, r->supp_channels[r->nscan]);

				r->nscan++;
				continue;
			}

			if (csmngr_radio_channel_is_unavailable(r, r->supp_channels[r->nscan])) {
				dbg("%s: Skip %u as it is 'unavailable' now\n",
				    r->ifname, r->supp_channels[r->nscan]);

				r->nscan++;
				continue;
			}

			if (r->scan_channel) {
				/* "scan-finished" event not received for the
				 * last triggered scan-channel.
				 * Get scanresults here anyway.
				 */
				ret = csmngr_radio_collect_channels_info(r);
				ret |= csmngr_radio_collect_scanresults(r);
				r->scan_channel = 0;
				if (ret)
					warn("%s: failed to get scanresults\n", r->ifname);
			}

			ret = csmngr_radio_scan(r, r->supp_channels[r->nscan]);
			if (!ret) {
				timer_set(&r->runtimer, r->scantimeout * 1000);
				r->nscan++;
				return 0;
			} else {
				//TODO: retry times
				dbg("%s: wifi scan failed (ret = %d). Retry later...\n", __func__, ret);
				timer_set(&r->runtimer, SCAN_FAIL_RETRY_INT * 1000);
				return 0;
			}
		};

		dbg("%s: Process scanres and survey data...\n", __func__);
		/* cycled through the scanlist */
		r->state = CS_STATE_PROCESS;
		r->nscan = 0;
		r->scan_channel = 0;
		timer_set(&r->runtimer, 500);
		break;

	case CS_STATE_PROCESS:
		ret = csmngr_process_survey_data(r);
		r->state = CS_STATE_IDLE;
		timer_set(&r->runtimer, 100);
		break;

	case CS_STATE_CAC:
		{
			uint32_t ch = 0;
			enum wifi_bw bw;

			ret = csmngr_radio_next_cac_channel(r, &ch, &bw);
			if (ret) {
				//TODO: cac-ongoing and/or missed cac-end event
				// read chinfo to know latest dfs-state
				dbg("%s: CS_STATE_CAC: cac-ongoing = %u\n",
				    __func__, r->cac_ongoing);
				break;
			}

			if (ch == 0) {
				dbg("%s: all DFS channels pre-cleared\n", r->ifname);
				r->state = CS_STATE_IDLE;
				break;
			}

			ret = csmngr_radio_trigger_cac(r, ch, bw);
			if (ret) {
				dbg("%s: CS_STATE_CAC: failed to start CAC\n", __func__);
				timer_set(&r->runtimer, CAC_FAIL_RETRY_INT * 1000);
			}
		}
		break;
	default:
		break;
	}

	return ret;
}

void csmngr_radio_runtimer_cb(atimer_t *t)
{
	struct radio_interface *r = container_of(t, struct radio_interface, runtimer);

	csmngr_radio_run(r);
}

int csmngr_start_engine(struct csmngr_private *priv)
{
	struct radio_interface *r = NULL;
	int ret = 0;

	list_for_each_entry(r, &priv->radiolist, list) {
		if (r->include) {
			r->refresh = r->mode == CS_MODE_ACS ? 1 : 0;
			ret |= csmngr_radio_run(r);
		}
	}

	return ret;
}

int csmngr_process_event_radio_scan_started(struct csmngr_private *priv,
					    struct radio_interface *r)
{
	dbg("%s: Received 'scan_started' event.\n", __func__);
	//TODO: start scan-timeout timer here

	return 0;
}

int csmngr_process_event_radio_scan_finished(struct csmngr_private *priv,
					     struct radio_interface *r)
{
	int ret = 0;

	dbg("%s: Received 'scan_finished' event.\n", __func__);
	/* scan-finished received in state != SCANNING is due to scan request
	 * not from us.
	 */
	if (r->state == CS_STATE_SCANNING) {
		if (timer_pending(&r->runtimer))
			timer_del(&r->runtimer);

		r->state = CS_STATE_IDLE;
		ret = csmngr_radio_collect_channels_info(r);
		ret |= csmngr_radio_collect_scanresults(r);
		timer_set(&r->runtimer, 1000);
	} else {
		ret = csmngr_radio_collect_channels_info(r);
		ret |= csmngr_radio_collect_scanresults(r);
	}

	r->scan_channel = 0;

	return ret;
}

int csmngr_process_event_radio_cac_start(struct csmngr_private *priv,
					 struct radio_interface *r,
					 uint32_t channel,
					 uint32_t bandwidth,
					 uint32_t cac_time)
{
	int ret = 0;

	dbg("%s: channel = %u, bw = %u, cac_time = %u\n", __func__,
	    channel, bandwidth, cac_time);

	r->cac_ongoing = true;
	if (r->state == CS_STATE_CAC) {
		timer_set(&r->runtimer, (cac_time + 5) * 1000);
		return 0;
	}

	ret = csmngr_radio_collect_channels_info(r);
	return ret;
}

int csmngr_process_event_radio_cac_end(struct csmngr_private *priv,
				       struct radio_interface *r,
				       uint32_t channel,
				       uint32_t bandwidth,
				       uint32_t status)
{
	int ret = 0;

	ret = csmngr_radio_collect_channels_info(r);
	r->cac_ongoing = false;

	/* cac-stop received in state != CAC is due to CAC not trigger by us */
	if (r->state == CS_STATE_CAC) {
		if (csmngr_radio_should_cac(r)) {
			timer_set(&r->runtimer, 1000);
		} else {
			r->state = CS_STATE_IDLE;

			/* all channels DFS cleared; perhaps a good time to
			 * re-evaluate channels.
			 */
			if (r->mode == CS_MODE_ACS) {
				r->refresh = 1;
				ret = csmngr_radio_run(r);
				timer_set(&r->runtimer, 500);
			}
		}
	}

	return ret;
}

int csmngr_process_event_radio_radar(struct csmngr_private *priv,
				     struct radio_interface *r,
				     uint32_t channel,
				     uint32_t bandwidth)
{
	int ret = 0;

	ret = csmngr_radio_collect_channels_info(r);

	//TODO
	return ret;
}
